# Copyright DataStax, Inc.
#
# Please see the included license file for details.
#
from six.moves import configparser
import importlib
import sys

from cassandra.auth import DSEPlainTextAuthProvider, DSEGSSAPIAuthProvider


def get_auth_provider(config_file, env, hostname, username=None, password=None, options=None):
    def print_debug(message):
        if options and hasattr(options, "debug") and options.debug:
            sys.stderr.write(message)

    if username:
        print_debug("Using DSEPlainTextAuthProvider\n")
        return DSEPlainTextAuthProvider(username=username, password=password)

    """
    Kerberos auth provider can be configured in the cqlshrc file [kerberos] section
    or with environment variables:

    Config option      Environment Variable      Description
    -------------      --------------------      -----------
    service            KRB_SERVICE               Service to authenticate with
    qops               QOPS                      Comma separated list of QOP values (default: auth)
    """
    configs = configparser.SafeConfigParser() if sys.version_info < (3, 2) else configparser.ConfigParser()
    configs.read(config_file)

    def get_option(section, option, env_variable, default=None):
        value = env.get(env_variable)
        if value is None:
            try:
                value = configs.get(section, option)
            except configparser.Error:
                value = default
        return value

    krb_service = get_option('kerberos', 'service', 'KRB_SERVICE', 'dse')
    krb_qop_value = get_option('kerberos', 'qops', 'QOPS', 'auth')

    try:
        provider_name = get_option('kerberos', 'dsecqlsh_auth_provider', 'DSECQLSH_AUTH_PROVIDER', 'DSEGSSAPIAuthProvider')
        try:
            # pluggable auth_providers must be placed in a `dselib/<name>.py` file
            cls = getattr(importlib.import_module('dselib.' + provider_name), provider_name)
        except:
            cls = globals()[provider_name] if provider_name in globals() else DSEGSSAPIAuthProvider
        provider = cls(service=krb_service, qops=krb_qop_value.split(','), hostname=hostname)
        print_debug("Using %s(service=%s, qops=%s, hostname=%s)\n" % (provider_name, krb_service, krb_qop_value, hostname))
        print_debug("    This will only be used if the server requests kerberos authentication\n")
        return provider
    except ImportError as e:
        print_debug("Attempted to use %s(service=%s, qops=%s, hostname=%s)\n" % (provider_name, krb_service, krb_qop_value, hostname))
        print_debug("    Attempt failed because: %s\n" % str(e))
        return None
